#include "paging.h"

#define PHYSICAL_MEMORY 0x4000000 // The size of physical memory, assume it is 64MB big.
#define PAGESIZE        0x1000    // 0x1000 = 4096 = 4K
#define XYZ_ADDRESS     0x300000 // ???

//TEST
int counter;
//TEST

extern ULONG kmalloc(ULONG sz);
extern ULONG kmalloc_a(ULONG sz);
extern ULONG kmalloc_ap(ULONG sz, ULONG *phys);
extern heap_t* kheap;
extern heap_t* create_heap(ULONG start, ULONG end_addr, ULONG max, UCHAR supervisor, UCHAR readonly);

page_directory_t* kernel_directory  = 0;
page_directory_t* current_directory = 0;

static void set_frame(ULONG frame_addr)
{
    ULONG frame = frame_addr/PAGESIZE;
    ULONG idx = INDEX_FROM_BIT(frame);
    ULONG off = OFFSET_FROM_BIT(frame);
    frames[idx] |= (0x1 << off);
}

static void clear_frame(ULONG frame_addr)
{
    ULONG frame = frame_addr/PAGESIZE;
    ULONG idx = INDEX_FROM_BIT(frame);
    ULONG off = OFFSET_FROM_BIT(frame);
    frames[idx] &= ~(0x1 << off);
}

static ULONG first_frame()
{
    ULONG i,j;
    for(i=0; i<INDEX_FROM_BIT(nframes); ++i)
    {
        if(frames[i] != 0xFFFFFFFF)
        {
            for(j=0; j<32; ++j)
            {
                ULONG toTest = 0x1 << j;
                if( !(frames[i] & toTest) )
                    return i*32+j;
            }
        }
    }
} //  warning: control reaches end of non-void function  // problem: return

void alloc_frame(page_t *page, int is_kernel, int is_writeable)
{
    if (page->frame != 0)
    {
        return;
    }
    else
    {
        ULONG idx = first_frame();
        if (idx == (ULONG)-1)
            printformat("alloc_frame: no free frames!!!");
        set_frame(idx*PAGESIZE);
        page->present = 1;
        page->rw    = ( is_writeable == 1 ) ? 1 : 0;
        page->user  = ( is_kernel    == 1 ) ? 0 : 1;
        page->frame = idx;
    }
}

void free_frame(page_t *page)
{
    ULONG frame;
    if (!(frame=page->frame))
    {
        return;
    }
    else
    {
        clear_frame(frame);
        page->frame = 0x0;
    }
}

void paging_install()
{
    ULONG mem_end_page = PHYSICAL_MEMORY;
    nframes = mem_end_page/PAGESIZE;
    frames  = (ULONG*)kmalloc(INDEX_FROM_BIT(nframes));
    k_memset(frames, 0, INDEX_FROM_BIT(nframes));

//TEST
    printformat("nframes: %d frames: %d \n", nframes, frames);
//TEST

    //ULONG phys;
    kernel_directory = (page_directory_t*)kmalloc_a(sizeof(page_directory_t));
    k_memset(kernel_directory, 0, sizeof(page_directory_t));
    kernel_directory->physicalAddr = (ULONG)kernel_directory->tablesPhysical;

//TEST
    printformat("kernel_directory->physicalAddr: %x \n", kernel_directory->physicalAddr);
//TEST
    int i = 0;
    for (i=KHEAP_START; i<KHEAP_START+KHEAP_INITIAL_SIZE; i+=PAGESIZE)
        get_page(i, 1, kernel_directory);

//TEST
    printformat("get_page \n");
//TEST

    i = 0;
    while (i < XYZ_ADDRESS ) //placement_address+0x1000)
    {
        alloc_frame( get_page(i, 1, kernel_directory), 0, 0);
        i += PAGESIZE;
    }
    for (i = KHEAP_START; i < KHEAP_START+KHEAP_INITIAL_SIZE; i+=PAGESIZE)
        alloc_frame( get_page(i, 1, kernel_directory), 0, 0);

//TEST
    printformat("alloc_frame part 1 and 2 \n");
//TEST

    // register_interrupt_handler(14, page_fault); // TODO???
    switch_page_directory(kernel_directory);//Hier liegt der Absturz begraben!

    kheap = create_heap(KHEAP_START, KHEAP_START+KHEAP_INITIAL_SIZE, 0xCFFFF000, 0, 0);
    current_directory = clone_directory(kernel_directory);
    switch_page_directory(current_directory);
}


page_t* get_page(ULONG address, int make, page_directory_t* dir)
{
    address /= PAGESIZE;
    ULONG table_idx = address / 1024;

    if (dir->tables[table_idx]) // If this table is already assigned
    {
        return &dir->tables[table_idx]->pages[address%1024];
    }
    else if(make)
    {
        ULONG tmp;
        dir->tables[table_idx] = (page_table_t*)kmalloc_ap(sizeof(page_table_t), &tmp);
        k_memset(dir->tables[table_idx], 0, PAGESIZE);
        dir->tablesPhysical[table_idx] = tmp | 0x7; // PRESENT, RW, US.
        return &dir->tables[table_idx]->pages[address%1024];
    }
    else
    {
        return 0;
    }
}

void switch_page_directory(page_directory_t* dir)
{
    current_directory = dir;
    asm volatile("mov %0, %%cr3":: "r"(dir->physicalAddr));

//TEST
    int counter;
    printformat("dir->physicalAddr: %x \n", dir->physicalAddr);
//TEST

    ULONG cr0;
    asm volatile("mov %%cr0, %0": "=r"(cr0));

//TEST
    printformat("switch_page_directory before paging switch \n");
    for (counter=0; counter<50000000; ++counter);
//TEST

    cr0 |= 0x80000000; // Enable paging by setting the PG bit in the CR0 register //Absturz!!!
    asm volatile("mov %0, %%cr0":: "r"(cr0));

//TEST
    printformat("switch_page_directory after paging switch \n");
    for (counter=0; counter<50000000; ++counter);
//TEST

}

void page_fault(registers_t* regs)
{
    ULONG faulting_address;
    asm volatile("mov %%cr2, %0" : "=r" (faulting_address));
    int present   = !(regs->err_code & 0x1); // Page not present
    int rw = regs->err_code & 0x2;           // Write operation?
    int us = regs->err_code & 0x4;           // Processor was in user-mode?
    int reserved = regs->err_code & 0x8;     // Overwritten CPU-reserved bits of page entry?
    //int id = regs->err_code & 0x10;          // Caused by an instruction fetch?

    printformat("Page fault! ( ");
    if (present) {printformat("present ");}
    if (rw) {printformat("read-only ");}
    if (us) {printformat("user-mode ");}
    if (reserved) {printformat("reserved ");}
    printformat(") at ");
    printformat("%x",faulting_address);
    printformat(" - EIP: ");
    printformat("%x",regs->eip);
    printformat("\n");
    printformat("Page fault");
}

static page_table_t *clone_table(page_table_t *src, ULONG *physAddr)
{
    page_table_t *table = (page_table_t*)kmalloc_ap(sizeof(page_table_t), physAddr);
    k_memset(table, 0, sizeof(page_directory_t));

    int i;
    for (i = 0; i < 1024; i++)
    {
        if (!src->pages[i].frame)
            continue;
        alloc_frame(&table->pages[i], 0, 0);
        if (src->pages[i].present) table->pages[i].present = 1;
        if (src->pages[i].rw)      table->pages[i].rw = 1;
        if (src->pages[i].user)    table->pages[i].user = 1;
        if (src->pages[i].accessed)table->pages[i].accessed = 1;
        if (src->pages[i].dirty)   table->pages[i].dirty = 1;

        // Physically copy the data across. This function is in process.s. //TODO!!!
        copy_page_physical(src->pages[i].frame*PAGESIZE, table->pages[i].frame*PAGESIZE);
    }
    return table;
}

page_directory_t *clone_directory(page_directory_t *src)
{
    ULONG phys;
    page_directory_t *dir = (page_directory_t*)kmalloc_ap(sizeof(page_directory_t), &phys);
    k_memset(dir, 0, sizeof(page_directory_t));
    ULONG offset = (ULONG)dir->tablesPhysical - (ULONG)dir;
    dir->physicalAddr = phys + offset;

    int i;
    for (i = 0; i < 1024; i++)
    {
        if (!src->tables[i])
            continue;

        if (kernel_directory->tables[i] == src->tables[i])
        {
            dir->tables[i] = src->tables[i];
            dir->tablesPhysical[i] = src->tablesPhysical[i];
        }
        else
        {
            ULONG phys;
            dir->tables[i] = clone_table(src->tables[i], &phys);
            dir->tablesPhysical[i] = phys | 0x07;
        }
    }
    return dir;
}


/* // not used function
static ULONG test_frame(ULONG frame_addr)
{
    ULONG frame = frame_addr/0x1000;
    ULONG idx = INDEX_FROM_BIT(frame);
    ULONG off = OFFSET_FROM_BIT(frame);
    return (frames[idx] & (0x1 << off));
}
*/
